@@ -12,6 +12,8 @@ use glob::glob;
 use yada::builder::DoubleArrayBuilder;
 use yada::DoubleArray;
 
+#[cfg(feature = "compress")]
+use lindera_compress::compress;
 use lindera_core::character_definition::{CharacterDefinitions, CharacterDefinitionsBuilder};
 use lindera_core::dictionary_builder::DictionaryBuilder;
 use lindera_core::error::LinderaErrorKind;
@@ -21,11 +23,15 @@ use lindera_core::unknown_dictionary::parse_unk;
 use lindera_core::user_dictionary::UserDictionary;
 use lindera_core::word_entry::{WordEntry, WordId};
 use lindera_core::LinderaResult;
+use lindera_decompress::Algorithm;
+
+const COMPRESS_ALGORITHM: Algorithm = Algorithm::LZMA { preset: 9 };
 
 #[derive(Debug)]
 pub struct CsvRow<'a> {
     surface_form: &'a str,
     left_id: u32,
+    #[allow(dead_code)]
     right_id: u32,
     word_cost: i32,
 
@@ -43,6 +49,9 @@ pub struct CsvRow<'a> {
 }
 
 impl<'a> CsvRow<'a> {
+    const SIMPLE_USERDIC_FIELDS_NUM: usize = 3;
+    const DETAILED_USERDIC_FIELDS_NUM: usize = 13;
+
     fn from_line(line: &'a str) -> LinderaResult<CsvRow<'a>> {
         let fields: Vec<_> = line.split(',').collect();
 
@@ -76,7 +85,7 @@ impl<'a> CsvRow<'a> {
         let fields: Vec<_> = line.split(',').collect();
 
         match fields.len() {
-            3 => Ok(CsvRow {
+            Self::SIMPLE_USERDIC_FIELDS_NUM => Ok(CsvRow {
                 surface_form: fields[0],
                 left_id: 0,
                 right_id: 0,
@@ -94,9 +103,11 @@ impl<'a> CsvRow<'a> {
                 reading: fields[2],
                 pronunciation: "*",
             }),
-            13 => CsvRow::from_line(line),
+            Self::DETAILED_USERDIC_FIELDS_NUM => CsvRow::from_line(line),
             _ => Err(LinderaErrorKind::Content.with_error(anyhow::anyhow!(
-                "user dictionary should be a CSV with 3 or 13 fields"
+                "user dictionary should be a CSV with {} or {} fields",
+                Self::SIMPLE_USERDIC_FIELDS_NUM,
+                Self::DETAILED_USERDIC_FIELDS_NUM
             ))),
         }
     }
@@ -105,11 +116,19 @@ impl<'a> CsvRow<'a> {
 pub struct IpadicBuilder {}
 
 impl IpadicBuilder {
+    const UNK_FIELDS_NUM: usize = 11;
+
     pub fn new() -> Self {
         IpadicBuilder {}
     }
 }
 
+impl Default for IpadicBuilder {
+    fn default() -> Self {
+        Self::new()
+    }
+}
+
 impl DictionaryBuilder for IpadicBuilder {
     fn build_dictionary(&self, input_dir: &Path, output_dir: &Path) -> LinderaResult<()> {
         fs::create_dir_all(&output_dir)
@@ -123,6 +142,32 @@ impl DictionaryBuilder for IpadicBuilder {
         Ok(())
     }
 
+    fn build_user_dictionary(&self, input_file: &Path, output_file: &Path) -> LinderaResult<()> {
+        let parent_dir = match output_file.parent() {
+            Some(parent_dir) => parent_dir,
+            None => {
+                return Err(LinderaErrorKind::Io.with_error(anyhow::anyhow!(
+                    "failed to get parent directory of output file"
+                )))
+            }
+        };
+        fs::create_dir_all(parent_dir)
+            .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
+
+        let user_dict = self.build_user_dict(input_file)?;
+
+        let mut wtr = io::BufWriter::new(
+            File::create(output_file)
+                .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?,
+        );
+        bincode::serialize_into(&mut wtr, &user_dict)
+            .map_err(|err| LinderaErrorKind::Serialize.with_error(anyhow::anyhow!(err)))?;
+        wtr.flush()
+            .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
+
+        Ok(())
+    }
+
     fn build_chardef(
         &self,
         input_dir: &Path,
@@ -135,14 +180,18 @@ impl DictionaryBuilder for IpadicBuilder {
         char_definitions_builder.parse(&char_def)?;
         let char_definitions = char_definitions_builder.build();
 
+        let mut chardef_buffer = Vec::new();
+        bincode::serialize_into(&mut chardef_buffer, &char_definitions)
+            .map_err(|err| LinderaErrorKind::Serialize.with_error(anyhow::anyhow!(err)))?;
+
         let wtr_chardef_path = output_dir.join(Path::new("char_def.bin"));
         println!("creating {:?}", wtr_chardef_path);
         let mut wtr_chardef = io::BufWriter::new(
             File::create(wtr_chardef_path)
                 .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?,
         );
-        bincode::serialize_into(&mut wtr_chardef, &char_definitions)
-            .map_err(|err| LinderaErrorKind::Serialize.with_error(anyhow::anyhow!(err)))?;
+        compress_write(&chardef_buffer, COMPRESS_ALGORITHM, &mut wtr_chardef)?;
+
         wtr_chardef
             .flush()
             .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
@@ -159,7 +208,11 @@ impl DictionaryBuilder for IpadicBuilder {
         println!("BUILD UNK");
         let unk_data_path = input_dir.join("unk.def");
         let unk_data = read_euc_file(&unk_data_path)?;
-        let unknown_dictionary = parse_unk(chardef.categories(), &unk_data)?;
+        let unknown_dictionary = parse_unk(chardef.categories(), &unk_data, Self::UNK_FIELDS_NUM)?;
+
+        let mut unk_buffer = Vec::new();
+        bincode::serialize_into(&mut unk_buffer, &unknown_dictionary)
+            .map_err(|err| LinderaErrorKind::Serialize.with_error(anyhow::anyhow!(err)))?;
 
         let wtr_unk_path = output_dir.join(Path::new("unk.bin"));
         println!("creating {:?}", wtr_unk_path);
@@ -167,8 +220,7 @@ impl DictionaryBuilder for IpadicBuilder {
             File::create(wtr_unk_path)
                 .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?,
         );
-        bincode::serialize_into(&mut wtr_unk, &unknown_dictionary)
-            .map_err(|err| LinderaErrorKind::Serialize.with_error(anyhow::anyhow!(err)))?;
+        compress_write(&unk_buffer, COMPRESS_ALGORITHM, &mut wtr_unk)?;
         wtr_unk
             .flush()
             .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
@@ -230,7 +282,7 @@ impl DictionaryBuilder for IpadicBuilder {
             .map(|line| CsvRow::from_line(line))
             .collect::<Result<_, _>>()
             .map_err(|err| LinderaErrorKind::Parse.with_error(anyhow::anyhow!(err)))?;
-        rows.sort_by_key(|row| row.surface_form.clone());
+        rows.sort_by_key(|row| row.surface_form);
 
         let wtr_da_path = output_dir.join(Path::new("dict.da"));
         println!("creating {:?}", wtr_da_path);
@@ -274,6 +326,8 @@ impl DictionaryBuilder for IpadicBuilder {
         );
 
         let mut words_buffer = Vec::new();
+        let mut words_idx_buffer = Vec::new();
+        println!("rows.len = {}", rows.len());
         for row in rows.iter() {
             let word = vec![
                 row.pos_level1.to_string(),
@@ -287,16 +341,16 @@ impl DictionaryBuilder for IpadicBuilder {
                 row.pronunciation.to_string(),
             ];
             let offset = words_buffer.len();
-            wtr_words_idx
+            words_idx_buffer
                 .write_u32::<LittleEndian>(offset as u32)
                 .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
             bincode::serialize_into(&mut words_buffer, &word)
                 .map_err(|err| LinderaErrorKind::Serialize.with_error(anyhow::anyhow!(err)))?;
         }
 
-        wtr_words
-            .write_all(&words_buffer[..])
-            .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
+        compress_write(&words_buffer, COMPRESS_ALGORITHM, &mut wtr_words)?;
+        compress_write(&words_idx_buffer, COMPRESS_ALGORITHM, &mut wtr_words_idx)?;
+
         wtr_words
             .flush()
             .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
@@ -325,17 +379,17 @@ impl DictionaryBuilder for IpadicBuilder {
             LinderaErrorKind::Io.with_error(anyhow::anyhow!("DoubleArray build error."))
         })?;
 
-        wtr_da
-            .write_all(&da_bytes[..])
-            .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
+        compress_write(&da_bytes, COMPRESS_ALGORITHM, &mut wtr_da)?;
 
+        let mut vals_buffer = Vec::new();
         for word_entries in word_entry_map.values() {
             for word_entry in word_entries {
                 word_entry
-                    .serialize(&mut wtr_vals)
+                    .serialize(&mut vals_buffer)
                     .map_err(|err| LinderaErrorKind::Serialize.with_error(anyhow::anyhow!(err)))?;
             }
         }
+        compress_write(&vals_buffer, COMPRESS_ALGORITHM, &mut wtr_vals)?;
         wtr_vals
             .flush()
             .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
@@ -363,7 +417,7 @@ impl DictionaryBuilder for IpadicBuilder {
         let forward_size = header[0] as u32;
         let backward_size = header[1] as u32;
         let len = 2 + (forward_size * backward_size) as usize;
-        let mut costs = vec![i16::max_value(); len];
+        let mut costs = vec![i16::MAX; len];
         costs[0] = forward_size as i16;
         costs[1] = backward_size as i16;
         for fields in lines_it {
@@ -379,11 +433,14 @@ impl DictionaryBuilder for IpadicBuilder {
             File::create(wtr_matrix_mtx_path)
                 .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?,
         );
+        let mut matrix_mtx_buffer = Vec::new();
         for cost in costs {
-            wtr_matrix_mtx
+            matrix_mtx_buffer
                 .write_i16::<LittleEndian>(cost)
                 .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
         }
+        compress_write(&matrix_mtx_buffer, COMPRESS_ALGORITHM, &mut wtr_matrix_mtx)?;
+
         wtr_matrix_mtx
             .flush()
             .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
@@ -402,7 +459,7 @@ impl DictionaryBuilder for IpadicBuilder {
             .map_err(|err| LinderaErrorKind::Parse.with_error(anyhow::anyhow!(err)))?;
 
         // sorting entries
-        rows.sort_by_key(|row| row.surface_form.clone());
+        rows.sort_by_key(|row| row.surface_form);
 
         let mut word_entry_map: BTreeMap<String, Vec<WordEntry>> = BTreeMap::new();
 
@@ -485,3 +542,30 @@ impl DictionaryBuilder for IpadicBuilder {
         })
     }
 }
+
+#[cfg(feature = "compress")]
+fn compress_write<W: Write>(
+    buffer: &[u8],
+    algorithm: Algorithm,
+    writer: &mut W,
+) -> LinderaResult<()> {
+    let compressed = compress(buffer, algorithm)
+        .map_err(|err| LinderaErrorKind::Compress.with_error(anyhow::anyhow!(err)))?;
+    bincode::serialize_into(writer, &compressed)
+        .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
+
+    Ok(())
+}
+
+#[cfg(not(feature = "compress"))]
+fn compress_write<W: Write>(
+    buffer: &[u8],
+    _algorithm: Algorithm,
+    writer: &mut W,
+) -> LinderaResult<()> {
+    writer
+        .write_all(buffer)
+        .map_err(|err| LinderaErrorKind::Io.with_error(anyhow::anyhow!(err)))?;
+
+    Ok(())
+}
